# THIS FILE IS GENERATED FROM macrosynergy SETUP.PY
short_version: str = '0.0.32'
version: str = '0.0.32'
full_version: str = '0.0.32'
git_revision: str = '16ec103208b508468ce8a9ddc1feb2b54145b321'
release: bool = True
if not release:
    version = full_version
